# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for conditional conditional."""

import unittest
from pyglove.core import symbolic
from pyglove.core import typing as pg_typing
from pyglove.core.patching import rule_based


class PatcherTest(unittest.TestCase):
  """Patcher test."""

  def assert_patch_equal(self, src, patcher_or_uri, expected_value):
    if isinstance(patcher_or_uri, rule_based.Patcher):
      patcher = patcher_or_uri
    else:
      patcher = rule_based.from_uri(patcher_or_uri)
    self.assertIsInstance(patcher, rule_based.Patcher)
    dst = patcher.patch(src)
    self.assertIs(src, dst)
    self.assertEqual(dst, expected_value)

  def test_patcher_without_typing(self):
    @rule_based.patcher()
    def set_value1(unused_src, k, v='0'):
      return {
          k: v
      }
    self.assert_patch_equal(symbolic.Dict(), set_value1(k='a'), {'a': '0'})  # pylint: disable=not-callable, no-value-for-parameter
    self.assert_patch_equal(symbolic.Dict(), 'set_value1?a&1', {'a': '1'})
    self.assert_patch_equal(symbolic.Dict(), 'set_value1?a&v=1', {'a': '1'})
    self.assert_patch_equal(symbolic.Dict(), 'set_value1?k=a&v=1', {'a': '1'})
    self.assertIn('set_value1', rule_based.patcher_names())

    with self.assertRaisesRegex(
        TypeError, 'The 1st argument of .* must be a symbolic type'):
      set_value1(k='a').patch(1)  # pylint: disable=not-callable, no-value-for-parameter

  def test_patcher_with_auto_typing(self):
    @rule_based.patcher(auto_typing=True)
    def set_value2(unused_src, k: str, v: int):
      return {
          k: v
      }
    self.assert_patch_equal(symbolic.Dict(), set_value2(k='a', v=1), {'a': 1})  # pylint: disable=no-value-for-parameter
    self.assert_patch_equal(symbolic.Dict(), 'set_value2?a&1', {'a': 1})
    self.assert_patch_equal(symbolic.Dict(), 'set_value2?a&v=1', {'a': 1})
    self.assert_patch_equal(symbolic.Dict(), 'set_value2?k=a&v=1', {'a': 1})
    self.assertIn('set_value2', rule_based.patcher_names())

  def test_patcher_with_typing(self):
    @rule_based.patcher([
        ('k', pg_typing.Str()),
        ('v', pg_typing.Int())
    ])
    def set_value2(unused_src, k, v):
      return {
          k: v
      }
    self.assert_patch_equal(symbolic.Dict(), set_value2(k='a', v=1), {'a': 1})  # pylint: disable=no-value-for-parameter
    self.assert_patch_equal(symbolic.Dict(), 'set_value2?a&1', {'a': 1})
    self.assert_patch_equal(symbolic.Dict(), 'set_value2?a&v=1', {'a': 1})
    self.assert_patch_equal(symbolic.Dict(), 'set_value2?k=a&v=1', {'a': 1})
    self.assertIn('set_value2', rule_based.patcher_names())

  def test_patcher_with_replacing_source_value(self):
    @rule_based.patcher([
        ('v', pg_typing.Int())
    ])
    def set_value3(unused_src, v):
      return {'': v}
    self.assertEqual(
        set_value3(v=1).patch(symbolic.Dict()), 1)   # pylint: disable=not-callable, no-value-for-parameter
    self.assertEqual(
        rule_based.from_uri('set_value3?1').patch(symbolic.Dict()), 1)
    self.assertIn('set_value3', rule_based.patcher_names())

  def test_patcher_with_returning_rebind_fn(self):
    @rule_based.patcher()
    def increment(unused_src):
      return lambda k, v: (v + 1) if isinstance(v, int) else v

    self.assertEqual(
        increment().patch(symbolic.Dict(a=1, b=2)), symbolic.Dict(a=2, b=3))  # pylint: disable=not-callable, no-value-for-parameter

    @rule_based.patcher()
    def no_change(unused_src):
      return lambda k, v: v

    self.assertEqual(
        no_change().patch(symbolic.Dict(a=1, b=2)), symbolic.Dict(a=1, b=2))  # pylint: disable=not-callable, no-value-for-parameter

  def test_patcher_with_composition(self):
    @rule_based.patcher([
        ('k', pg_typing.Str()),
        ('v', pg_typing.Int())
    ])
    def set_value2(unused_src, k, v):
      return {
          k: v
      }

    @rule_based.patcher()
    def compound(unused_src):
      return [
          set_value2(k='a', v=1),   # pylint: disable=no-value-for-parameter
          'set_value2?k=b&v=2',
          {
              'c': 3
          }
      ]
    self.assertEqual(
        compound().patch(symbolic.Dict(x=1)), symbolic.Dict(x=1, a=1, b=2, c=3))  # pylint: disable=no-value-for-parameter

  def test_patcher_with_validation(self):
    @rule_based.patcher()
    def set_value_with_validate(unused_src):
      def validate(x):
        if 'a' not in x:
          raise ValueError('Key `a` does not exist.')
      return {'a': 1}, validate

    p = set_value_with_validate()  # pylint: disable=not-callable, no-value-for-parameter
    x = p.patch(symbolic.Dict())
    p.validate(x)
    del x['a']
    with self.assertRaisesRegex(
        ValueError, 'Key `a` does not exist.'):
      p.validate(x)

  def test_bad_patcher(self):
    @rule_based.patcher()
    def set_value4(unused_src):
      return 1

    with self.assertRaisesRegex(
        TypeError,
        'Patching rule .* should be a dict of path to values'):
      set_value4().patch(symbolic.Dict())  # pylint: disable=not-callable, no-value-for-parameter

    # Bad patcher argument spec.
    with self.assertRaisesRegex(
        TypeError,
        '.* cannot be used for constraining Patcher target.'):
      @rule_based.patcher([
          ('unused_src', pg_typing.Int()),
      ])   # pylint: disable=unused-variable
      def set_value5(unused_src, x):
        del x

    # Bad patcher argument spec.
    with self.assertRaisesRegex(
        TypeError,
        '.* cannot be used for constraining Patcher argument'):
      @rule_based.patcher([
          ('x', pg_typing.Enum(1, [1, 2]))
      ])   # pylint: disable=unused-variable
      def set_value6(unused_src, x=1):
        del x

    # Bad patcher without target argument.
    with self.assertRaisesRegex(
        TypeError,
        r'Patcher function should have at least 1 argument as patching target'):
      @rule_based.patcher()
      def set_value7():  # pylint: disable=unused-variable
        return {'': 1}

    # Bad patcher with wrong return type for validator.
    @rule_based.patcher()
    def set_value8(unused_src):
      return {'': 1}, 1

    with self.assertRaisesRegex(
        TypeError,
        r'The validator returned from patcher .* is not callable'):
      set_value8().patch(symbolic.Dict())  # pylint: disable=no-value-for-parameter


@rule_based.patcher([
    ('v', pg_typing.Int())
])
def p1(unused_src, k, v):
  return {k: v}


@rule_based.patcher()
def p2(unused_src):  # pylint: disable=unused-variable
  return {'': symbolic.Dict()}


@rule_based.patcher()
def p3(unused_src):
  def validate(x):
    if 'a' not in x:
      raise ValueError('Key `a` does not exist.')
  return {}, validate


class PatchTest(unittest.TestCase):
  """Tests for `pg.patch`."""

  def test_patch_with_a_single_rule(self):
    self.assertEqual(
        rule_based.patch(symbolic.Dict(x=1, y=2), {'x': 0, 'z': 3}),
        symbolic.Dict(x=0, y=2, z=3))

  def test_patch_with_multiple_rules_without_replacement(self):
    self.assertEqual(
        rule_based.patch(symbolic.Dict(), [
            p1(k='a', v=1),  # pylint: disable=no-value-for-parameter
            'p1?b&2',
            'p1?c&v=3',
            'p1?k=d&v=4',
            'p1?a&2'
        ]),
        {
            'a': 2,
            'b': 2,
            'c': 3,
            'd': 4
        })

  def test_patch_with_multiple_rules_with_replacement(self):
    self.assertEqual(
        rule_based.patch(symbolic.Dict(), [
            p1(k='a', v=1),  # pylint: disable=no-value-for-parameter
            'p1?b&2',
            'p1?c&v=3',
            'p2',
            'p1?k=d&v=4',
            'p1?a&2'
        ]),
        {
            'a': 2,
            'd': 4
        })

  def test_patch_with_a_rebind_dict(self):
    self.assertEqual(
        rule_based.patch(symbolic.Dict(a=1, b=2), {'a': 0, 'c': 3}),
        symbolic.Dict(a=0, b=2, c=3))

  def test_patch_with_a_rebind_function(self):
    self.assertEqual(
        rule_based.patch(
            symbolic.Dict(a=1),
            lambda k, v, p: (v + 1) if isinstance(v, int) else v),
        symbolic.Dict(a=2))

    self.assertEqual(
        rule_based.patch(symbolic.Dict(a=1), [
            lambda k, v: v
        ]),
        symbolic.Dict(a=1))

  def test_patch_with_validator(self):
    self.assertEqual(
        rule_based.patch(symbolic.Dict(), [
            p1(k='a', v=1),  # pylint: disable=no-value-for-parameter
            'p1?b&2',
            'p3'
        ]),
        {
            'a': 1,
            'b': 2,
        })

  def test_nothing_to_patch(self):
    self.assertEqual(rule_based.patch(symbolic.Dict(), []), {})

  def test_bad_patch(self):
    with self.assertRaisesRegex(ValueError, 'Key `a` does not exist.'):
      _ = rule_based.patch(symbolic.Dict(), [
          'p1?b&2',
          p3()  # pylint: disable=no-value-for-parameter
      ])

    # Bad patch input.
    with self.assertRaisesRegex(
        TypeError, 'Patching rule .* should be a dict of path to values'):
      rule_based.patch(symbolic.Dict(), [1])


class PatcherHelpersTest(unittest.TestCase):
  """Tests for patcher helpers."""

  def test_is_patcher_parameter_spec(self):
    def assert_argument_spec(value_spec):
      self.assertTrue(rule_based._is_patcher_parameter_spec(value_spec))

    def assert_not_argument_spec(value_spec):
      self.assertFalse(rule_based._is_patcher_parameter_spec(value_spec))

    assert_argument_spec(pg_typing.Any())
    assert_argument_spec(pg_typing.Int())
    assert_argument_spec(pg_typing.Float())
    assert_argument_spec(pg_typing.Bool())
    assert_argument_spec(pg_typing.Str())
    assert_argument_spec(pg_typing.Enum('a', ['a', 'b']))
    assert_argument_spec(pg_typing.Enum('a', ['a', 'b']))
    assert_argument_spec(pg_typing.List(pg_typing.Int()))
    assert_argument_spec(pg_typing.List(pg_typing.Enum('a', ['a', 'b'])))

    assert_not_argument_spec(pg_typing.Dict())
    assert_not_argument_spec(pg_typing.Type(Exception))
    assert_not_argument_spec(pg_typing.Object(ValueError))
    assert_not_argument_spec(pg_typing.Enum(1, [0, 1]))
    assert_not_argument_spec(pg_typing.List(pg_typing.List(pg_typing.Int())))
    assert_not_argument_spec(
        pg_typing.Union([pg_typing.Int(), pg_typing.Float()]))

  def test_from_uri(self):
    @rule_based.patcher([
        ('x', pg_typing.Int())
    ], name='my_patcher')
    def foo(src, x, y='bar'):
      del src, x, y

    self.assertEqual(rule_based.from_uri('my_patcher?1'), foo(x=1))    # pylint: disable=not-callable, no-value-for-parameter
    self.assertEqual(rule_based.from_uri('my_patcher?x=1'), foo(x=1))  # pylint: disable=not-callable, no-value-for-parameter

    with self.assertRaisesRegex(
        KeyError, 'Patcher .* is not registered'):
      rule_based.from_uri('foo')

    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to int'
        r'.*Patcher=.*foo\', Argument=\'x\''):
      rule_based.from_uri('my_patcher?x=a').patch(symbolic.Dict())

    with self.assertRaisesRegex(
        KeyError,
        r'Too many positional arguments are provided.'):
      rule_based.from_uri('my_patcher?1&2&3').patch(symbolic.Dict())

    with self.assertRaisesRegex(
        TypeError,
        r'foo\(\) (missing 1 required positional argument|'
        'takes at least 2 arguments)'):
      rule_based.from_uri('my_patcher').patch(symbolic.Dict())

  def test_parse_uri(self):
    def test_parse(uri, expected_name, expected_args, expected_kwargs):
      name, args, kwargs = rule_based.parse_uri(uri)
      self.assertEqual(name, expected_name)
      self.assertEqual(args, expected_args)
      self.assertEqual(kwargs, expected_kwargs)

    test_parse('foo', 'foo', [], {})
    test_parse('foo?', 'foo', [''], {})
    test_parse('foo?1', 'foo', ['1'], {})
    test_parse('foo?1&a:b&c', 'foo', ['1', 'a:b', 'c'], {})
    test_parse('foo?a=1', 'foo', [], {'a': '1'})
    test_parse('foo?a=1&b=2', 'foo', [], {'a': '1', 'b': '2'})
    test_parse('foo?1:2:3&b=2', 'foo', ['1:2:3'], {'b': '2'})

    with self.assertRaisesRegex(
        ValueError,
        '.* is not a valid Patcher name.'):
      rule_based.parse_uri('1foo')

    with self.assertRaisesRegex(
        ValueError,
        'Invalid argument specification: a-1=2'):
      rule_based.parse_uri('foo?a-1=2')

    with self.assertRaisesRegex(
        ValueError,
        'Positional argument should be provided before keyword arguments'):
      rule_based.parse_uri('foo?a=1&2')

  def test_parse_args(self):
    signature = pg_typing.Signature(
        pg_typing.CallableType.FUNCTION, 'foo', '__main__',
        [
            pg_typing.Argument(
                'src',
                pg_typing.Argument.Kind.POSITIONAL_OR_KEYWORD,
                pg_typing.Any()
            ),
            pg_typing.Argument(
                'x',
                pg_typing.Argument.Kind.POSITIONAL_OR_KEYWORD,
                pg_typing.Int()
            ),
            pg_typing.Argument(
                'y',
                pg_typing.Argument.Kind.POSITIONAL_OR_KEYWORD,
                pg_typing.List(pg_typing.Float(min_value=0.0, max_value=1.0))
            )
        ])
    args, kwargs = rule_based.parse_args(signature, ['0'], {'y': '0.1:0.5'})
    self.assertEqual(args, [0])
    self.assertEqual(kwargs, {'y': [0.1, 0.5]})

    with self.assertRaisesRegex(
        KeyError, 'Unexpected argument'):
      rule_based.parse_args(signature, [], {'z': '0.1'})

  def test_allow_repeated_patcher_registration(self):
    """Test `rule_based.allow_repeated_patcher_registration`."""
    @rule_based.patcher(name='my_patch')
    def patcher1(unused_src):
      return {}

    @rule_based.patcher(name='my_patch')
    def patcher2(unused_src):
      return {}

    rule_based.allow_repeated_patcher_registration(False)
    with self.assertRaisesRegex(
        KeyError, 'Patcher .* already registered.'):
      @rule_based.patcher(name='my_patch')
      def patcher3(unused_src):
        return {}
    rule_based.allow_repeated_patcher_registration(True)


def parse_arg(value_spec, arg_str):
  return rule_based.parse_arg('__main__.patcher_a', 'x', value_spec, arg_str)


class ParseArgTest(unittest.TestCase):

  def test_parse_float(self):
    self.assertTrue(parse_arg(pg_typing.Bool(), 'True'))
    self.assertTrue(parse_arg(pg_typing.Bool(), 'true'))
    self.assertTrue(parse_arg(pg_typing.Bool(), 'yes'))
    self.assertTrue(parse_arg(pg_typing.Bool(), '1'))
    self.assertFalse(parse_arg(pg_typing.Bool(), 'False'))
    self.assertFalse(parse_arg(pg_typing.Bool(), 'false'))
    self.assertFalse(parse_arg(pg_typing.Bool(), 'no'))
    self.assertFalse(parse_arg(pg_typing.Bool(), '0'))
    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to bool. '
        r'\(Patcher=\'__main__.patcher_a\', Argument=\'x\'\)'):
      parse_arg(pg_typing.Bool(), 'a')

    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to float. '
        r'\(Patcher=\'__main__.patcher_a\', Argument=\'x\'\)'):
      parse_arg(pg_typing.Float(), 'a')

    with self.assertRaisesRegex(
        ValueError,
        'Value -1.0 is out of range .*path=__main__.patcher_a.x'):
      parse_arg(pg_typing.Float(min_value=0.), '-1.0')

  def test_parse_int(self):
    self.assertEqual(parse_arg(pg_typing.Int(), '1'), 1)
    self.assertIsNone(parse_arg(pg_typing.Int().noneable(), 'None'))
    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to int. '
        r'\(Patcher=\'__main__.patcher_a\', Argument=\'x\'\)'):
      parse_arg(pg_typing.Int(), 'a')

    with self.assertRaisesRegex(
        ValueError,
        'Value -1 is out of range .*path=__main__.patcher_a.x'):
      parse_arg(pg_typing.Int(min_value=0), '-1')

    # Test float.
    self.assertEqual(parse_arg(pg_typing.Float(), '1'), 1.0)
    self.assertIsNone(parse_arg(pg_typing.Float().noneable(), 'None'))
    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to float. '
        r'\(Patcher=\'__main__.patcher_a\', Argument=\'x\'\)'):
      parse_arg(pg_typing.Float(), 'a')
    with self.assertRaisesRegex(
        ValueError, 'Value -1.0 is out of range'):
      parse_arg(pg_typing.Float(min_value=0.), '-1.0')

  def test_parse_str(self):
    self.assertEqual(parse_arg(pg_typing.Str(), '1'), '1')
    self.assertEqual(parse_arg(pg_typing.Str(), '"1"'), '1')
    self.assertIsNone(parse_arg(pg_typing.Str().noneable(), 'none'))
    self.assertEqual(parse_arg(pg_typing.Str().noneable(), '"none"'), 'none')

    with self.assertRaisesRegex(
        ValueError,
        'String \'bar\' does not match regular expression '
        '.*path=__main__.patcher_a.x'):
      parse_arg(pg_typing.Str(regex='foo.*'), 'bar')

    with self.assertRaisesRegex(
        ValueError,
        'Unmatched quote for string value: \'"1\'.*'):
      parse_arg(pg_typing.Str(), '"1')

    with self.assertRaisesRegex(
        ValueError,
        'Unmatched quote for string value: \'1"\'.*'):
      parse_arg(pg_typing.Str(), '1"')

  def test_parse_enum(self):
    self.assertEqual(parse_arg(pg_typing.Enum('a', ['a', 'b']), 'b'), 'b')
    with self.assertRaisesRegex(
        ValueError,
        r'.* cannot be used for Patcher argument. '
        r'Only Enum of string type can be used.'):
      parse_arg(pg_typing.Enum(1, [1, 2]), '1')

  def test_parse_list(self):
    self.assertEqual(parse_arg(pg_typing.List(pg_typing.Int()), ''), [])
    self.assertEqual(parse_arg(pg_typing.List(pg_typing.Int()), '1'), [1])
    self.assertEqual(parse_arg(pg_typing.List(pg_typing.Int()), '1:2'),
                     [1, 2])
    self.assertEqual(parse_arg(pg_typing.List(pg_typing.Bool()), 'yes:no'),
                     [True, False])
    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to int. '
        r'\(Patcher=\'__main__.patcher_a\', Argument=\'x\[0\]\'\)'):
      parse_arg(pg_typing.List(pg_typing.Int()), 'a:b')

  def test_parse_unsupported_types(self):
    with self.assertRaisesRegex(
        ValueError,
        r'Dict\(\) cannot be used for Patcher argument'):
      parse_arg(pg_typing.Dict(), '{}')


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.symbolize."""

import unittest

from pyglove.core import typing as pg_typing
from pyglove.core.symbolic.base import eq as pg_eq
from pyglove.core.symbolic.base import from_json as pg_from_json
from pyglove.core.symbolic.class_wrapper import ClassWrapper
from pyglove.core.symbolic.dict import Dict
from pyglove.core.symbolic.functor import Functor
from pyglove.core.symbolic.list import List
from pyglove.core.symbolic.object import Object
from pyglove.core.symbolic.symbolize import symbolize as pg_symbolize


class SymbolizeRegularTypesTest(unittest.TestCase):
  """Tests for `pg.symbolize`."""

  def test_symboliz_dict(self):
    self.assertIs(pg_symbolize(dict), Dict)
    with self.assertRaisesRegex(
        ValueError, 'Constraints are not supported in symbolic'):
      pg_symbolize(dict, pg_typing.Dict())

  def test_symbolize_list(self):
    self.assertIs(pg_symbolize(list), List)
    with self.assertRaisesRegex(
        ValueError, 'Constraints are not supported in symbolic'):
      pg_symbolize(list, pg_typing.List(pg_typing.Any()))


class SymbolizeFunctionsTest(unittest.TestCase):
  """Tests for symbolizing functions."""

  def test_symbolize_a_function_by_function_call(self):
    def f(x, y):
      del x, y
    self.assertTrue(issubclass(pg_symbolize(f), Functor))

  def test_symbolize_a_function_by_decorator_without_call(self):
    @pg_symbolize
    def f(x, y):
      del x, y
    self.assertTrue(issubclass(f, Functor))

  def test_symbolize_a_function_by_decorator_without_args(self):
    @pg_symbolize()
    def f(x, y):
      del x, y
    self.assertTrue(issubclass(f, Functor))

  def test_symbolize_a_function_by_decorator_with_typing(self):
    @pg_symbolize([
        ('x', pg_typing.Int()),
        ('y', pg_typing.Str())
    ], returns=pg_typing.Int())
    def f(x, y):
      del x, y
    self.assertTrue(issubclass(f, Functor))
    self.assertEqual(
        f.__signature__.args,
        [
            pg_typing.Argument(
                'x',
                pg_typing.Argument.Kind.POSITIONAL_OR_KEYWORD,
                pg_typing.Int()
            ),
            pg_typing.Argument(
                'y',
                pg_typing.Argument.Kind.POSITIONAL_OR_KEYWORD,
                pg_typing.Str()
            ),
        ],
    )
    self.assertEqual(f.__signature__.return_value, pg_typing.Int())

  def test_symbolize_with_serialization_key(self):
    @pg_symbolize(serialization_key='BAR', additional_keys=['RRR'])
    def bar(a, b):
      del a, b

    b = bar(3, 4)
    v = b.to_json()
    self.assertEqual(v['_type'], 'BAR')

    # Deserialize with type name: OK
    v['_type'] = bar.__type_name__
    self.assertTrue(pg_eq(pg_from_json(v), b))

    # Deserialize with additional key: OK
    v['_type'] = 'RRR'
    self.assertTrue(pg_eq(pg_from_json(v), b))


class SymbolizeClassesTest(unittest.TestCase):
  """Tests for symbolizing classes.

  For more detailed cases, see 'functor_test.py'.
  """

  def test_symbolize_a_class_by_function_call(self):

    class A:
      def __init__(self, x, y):
        pass

    A1 = pg_symbolize(A)  # pylint: disable=invalid-name
    self.assertIsInstance(A1(1, 2), A)
    self.assertIsInstance(A1(1, 2), ClassWrapper)
    self.assertEqual(list(A1.__schema__.fields.keys()), ['x', 'y'])
    self.assertEqual(
        [f.value for f in A1.__schema__.fields.values()],
        [pg_typing.Any(), pg_typing.Any()],
    )

  def test_symbolize_a_class_by_decorator_without_call(self):

    @pg_symbolize
    class A:
      def __init__(self, x, y):
        pass

    self.assertIsInstance(A(1, 2), ClassWrapper)
    self.assertEqual(list(A.__schema__.fields.keys()), ['x', 'y'])
    self.assertEqual(
        [f.value for f in A.__schema__.fields.values()],
        [pg_typing.Any(), pg_typing.Any()],
    )

  def test_symbolize_a_class_by_decorator_without_args(self):

    @pg_symbolize()
    class A:
      def __init__(self, x, y):
        pass

    self.assertIsInstance(A(1, 2), ClassWrapper)
    self.assertEqual(list(A.__schema__.fields.keys()), ['x', 'y'])
    self.assertEqual(
        [f.value for f in A.__schema__.fields.values()],
        [pg_typing.Any(), pg_typing.Any()],
    )

  def test_symbolize_a_class_by_decorator_with_typing(self):

    @pg_symbolize([
        ('x', pg_typing.Int()),
        ('y', pg_typing.Str()),
    ])
    class A:
      def __init__(self, x, y):
        pass

    self.assertEqual(list(A.__schema__.fields.keys()), ['x', 'y'])
    self.assertEqual(
        [f.value for f in A.__schema__.fields.values()],
        [pg_typing.Int(), pg_typing.Str()],
    )

  def test_symbolize_pg_object_subclass(self):

    class A(Object):
      pass

    with self.assertRaisesRegex(
        ValueError,
        'Cannot symbolize .* is already a dataclass-like symbolic class'):
      pg_symbolize(A)

  def test_symbolize_an_existing_symbolized_class(self):

    @pg_symbolize
    class A:
      pass

    self.assertIs(pg_symbolize(A), A)

  def test_symbolize_with_serialization_key(self):

    @pg_symbolize(serialization_key='BB', additional_keys=['BBB'])
    class B:

      def __init__(self, x):
        self.x = x

    b = B(1)
    v = b.to_json()
    self.assertEqual(v['_type'], 'BB')

    # Deserilaize with serialization key: OK
    self.assertTrue(pg_eq(pg_from_json(v), b))

    # Deserialize with type name: OK
    v['_type'] = B.__type_name__
    self.assertTrue(pg_eq(pg_from_json(v), b))

    # Deserialize with additional key: OK
    v['_type'] = 'BBB'
    self.assertTrue(pg_eq(pg_from_json(v), b))

  def test_bad_symbolize(self):

    class A:
      pass

    with self.assertRaisesRegex(
        ValueError, 'Only `constraint` is supported as positional arguments'):
      pg_symbolize(A, 1, 2)

    with self.assertRaisesRegex(TypeError, '.* cannot be symbolized'):
      pg_symbolize(1)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import unittest
from pyglove.core import monitoring
from pyglove.core.symbolic import error_info  # pylint: disable=unused-import


class MetricCollectionTest(unittest.TestCase):
  """Tests for metric collection."""

  def test_default_metric_collection_cls(self):
    self.assertIs(
        monitoring.default_metric_collection_cls(),
        monitoring.InMemoryMetricCollection
    )

    class TestMetricCollection(monitoring.MetricCollection):
      pass

    monitoring.set_default_metric_collection_cls(TestMetricCollection)
    self.assertIs(
        monitoring.default_metric_collection_cls(),
        TestMetricCollection
    )
    monitoring.set_default_metric_collection_cls(
        monitoring.InMemoryMetricCollection
    )
    self.assertIs(
        monitoring.default_metric_collection_cls(),
        monitoring.InMemoryMetricCollection
    )

  def test_metric_collection(self):
    collection = monitoring.metric_collection('/test')
    self.assertEqual(collection.namespace, '/test')
    self.assertIsInstance(collection, monitoring.InMemoryMetricCollection)

  def test_creation_failures(self):
    collection = monitoring.InMemoryMetricCollection('/test')
    counter = collection.get_counter('counter', 'counter description')
    self.assertIsInstance(counter, monitoring.Counter)
    with self.assertRaisesRegex(
        ValueError, 'Metric .* already exists with a different type'
    ):
      collection.get_distribution('counter', 'counter description')

    with self.assertRaisesRegex(
        ValueError, 'Metric .* already exists with a different description'
    ):
      collection.get_counter('counter', 'different description')

    with self.assertRaisesRegex(
        ValueError,
        'Metric .* already exists with different parameter definitions'
    ):
      collection.get_counter(
          'counter', 'counter description', parameters={'field1': str}
      )


class InMemoryDistributionValueTest(unittest.TestCase):
  """Tests for in memory distribution value."""

  def test_empty_distribution(self):
    dist = monitoring._InMemoryDistributionValue()
    self.assertEqual(dist.count, 0)
    self.assertEqual(dist.sum, 0.0)
    self.assertEqual(dist.mean, 0.0)
    self.assertEqual(dist.stddev, 0.0)
    self.assertEqual(dist.variance, 0.0)
    self.assertEqual(dist.median, 0.0)
    self.assertEqual(dist.percentile(50), 0.0)
    self.assertEqual(dist.fraction_less_than(100), 0.0)

  def test_add_value(self):
    dist = monitoring._InMemoryDistributionValue()
    dist.add(1)
    dist.add(3)
    dist.add(10)
    dist.add(2)
    self.assertEqual(dist.count, 4)
    self.assertEqual(dist.sum, 16.0)
    self.assertEqual(dist.mean, 4.0)
    self.assertEqual(dist.stddev, 3.5355339059327378)
    self.assertEqual(dist.variance, 12.5)
    self.assertEqual(dist.median, 2.5)
    self.assertEqual(dist.percentile(50), 2.5)
    self.assertEqual(dist.percentile(10), 1.3)
    self.assertEqual(dist.fraction_less_than(100), 1.0)
    self.assertEqual(dist.fraction_less_than(1), 0.0)
    self.assertEqual(dist.fraction_less_than(10), 0.75)

  def test_add_value_no_numpy(self):
    numpy = monitoring.numpy
    monitoring.numpy = None
    dist = monitoring._InMemoryDistributionValue()
    dist.add(1)
    dist.add(3)
    dist.add(10)
    dist.add(2)
    self.assertEqual(dist.count, 4)
    self.assertEqual(dist.sum, 16.0)
    self.assertEqual(dist.mean, 4.0)
    self.assertEqual(dist.stddev, 3.5355339059327378)
    self.assertEqual(dist.variance, 12.5)
    self.assertEqual(dist.median, 2.5)
    self.assertEqual(dist.percentile(50), 2.5)
    self.assertEqual(dist.percentile(10), 1.3)
    self.assertEqual(dist.fraction_less_than(100), 1.0)
    self.assertEqual(dist.fraction_less_than(1), 0.0)
    self.assertEqual(dist.fraction_less_than(10), 0.75)
    monitoring.numpy = numpy

  def test_window_size(self):
    dist = monitoring._InMemoryDistributionValue(window_size=3)
    dist.add(1)
    dist.add(3)
    dist.add(10)
    dist.add(2)
    self.assertEqual(dist.count, 3)
    self.assertEqual(dist.sum, 15.0)
    self.assertEqual(dist.mean, 5.0)
    self.assertEqual(dist.stddev, 3.5590260840104366)
    self.assertEqual(dist.variance, 12.666666666666664)
    self.assertEqual(dist.median, 3.0)
    self.assertEqual(dist.percentile(50), 3.0)
    self.assertEqual(dist.percentile(10), 2.2)
    self.assertEqual(dist.fraction_less_than(100), 1.0)
    self.assertEqual(dist.fraction_less_than(1), 0.0)
    self.assertEqual(dist.fraction_less_than(10), 0.6666666666666666)


class InMemoryCounterTest(unittest.TestCase):
  """Tests for in memory counter."""

  def test_counter_without_parameters(self):
    collection = monitoring.InMemoryMetricCollection('/test')
    counter = collection.get_counter('counter', 'counter description')
    self.assertEqual(counter.namespace, '/test')
    self.assertEqual(counter.name, 'counter')
    self.assertEqual(counter.description, 'counter description')
    self.assertEqual(counter.parameter_definitions, {})
    self.assertEqual(counter.full_name, '/test/counter')
    self.assertEqual(counter.value(), 0)
    self.assertEqual(counter.increment(), 1)
    self.assertEqual(counter.value(), 1)
    self.assertEqual(counter.increment(2), 3)
    self.assertEqual(counter.value(), 3)
    self.assertIs(collection.metrics()[0], counter)

  def test_counter_with_parameters(self):
    collection = monitoring.InMemoryMetricCollection('/test')
    counter = collection.get_counter(
        'counter', 'counter description', {'field1': str}
    )
    self.assertEqual(counter.namespace, '/test')
    self.assertEqual(counter.name, 'counter')
    self.assertEqual(counter.description, 'counter description')
    self.assertEqual(counter.parameter_definitions, {'field1': str})
    self.assertEqual(counter.full_name, '/test/counter')
    self.assertEqual(counter.value(field1='foo'), 0)
    self.assertEqual(counter.increment(field1='foo'), 1)
    self.assertEqual(counter.value(field1='bar'), 0)
    self.assertEqual(counter.increment(field1='bar'), 1)
    self.assertEqual(counter.increment(field1='foo', delta=2), 3)
    self.assertEqual(counter.value(field1='foo'), 3)

    with self.assertRaisesRegex(TypeError, '.* has type .* but expected type'):
      counter.increment(field1=1)

    with self.assertRaisesRegex(KeyError, '.* is required but not given'):
      counter.increment()

    with self.assertRaisesRegex(KeyError, '.* is not defined but provided'):
      counter.increment(field1='foo', field2='a')


class InMemoryScalarTest(unittest.TestCase):
  """Tests for in memory scalar."""

  def test_scalar_without_parameters(self):
    collection = monitoring.InMemoryMetricCollection('/test')
    scalar = collection.get_scalar('scalar', 'scalar description')
    self.assertEqual(scalar.namespace, '/test')
    self.assertEqual(scalar.name, 'scalar')
    self.assertEqual(scalar.description, 'scalar description')
    self.assertEqual(scalar.parameter_definitions, {})
    self.assertEqual(scalar.full_name, '/test/scalar')
    self.assertEqual(scalar.value(), 0)
    self.assertEqual(scalar.increment(), 1)
    self.assertEqual(scalar.value(), 1)
    scalar.set(3)
    self.assertEqual(scalar.increment(2), 5)
    self.assertEqual(scalar.value(), 5)

  def test_scalar_with_parameters(self):
    collection = monitoring.InMemoryMetricCollection('/test')
    scalar = collection.get_scalar(
        'scalar', 'scalar description', {'field1': str}, float
    )
    self.assertEqual(scalar.namespace, '/test')
    self.assertEqual(scalar.name, 'scalar')
    self.assertEqual(scalar.description, 'scalar description')
    self.assertEqual(scalar.parameter_definitions, {'field1': str})
    self.assertEqual(scalar.full_name, '/test/scalar')
    self.assertEqual(scalar.value(field1='foo'), 0.0)
    scalar.set(2.5, field1='bar')
    self.assertEqual(scalar.value(field1='bar'), 2.5)
    self.assertEqual(scalar.increment(1.1, field1='bar'), 3.6)
    self.assertEqual(scalar.value(field1='bar'), 3.6)
    self.assertEqual(scalar.value(field1='foo'), 0.0)


class InMemoryDistributionTest(unittest.TestCase):
  """Tests for in memory distribution."""

  def test_distribution_without_parameters(self):
    collection = monitoring.InMemoryMetricCollection('/test')
    dist = collection.get_distribution(
        'distribution', 'distribution description'
    )
    self.assertEqual(dist.namespace, '/test')
    self.assertEqual(dist.name, 'distribution')
    self.assertEqual(dist.description, 'distribution description')
    self.assertEqual(dist.parameter_definitions, {})
    self.assertEqual(dist.full_name, '/test/distribution')
    v = dist.value()
    self.assertEqual(v.count, 0)
    dist.record(1)
    dist.record(2)
    dist.record(3)
    v = dist.value()
    self.assertEqual(v.count, 3)

    dist = collection.get_distribution(
        'distribution2', 'distribution description'
    )
    with dist.record_duration():
      time.sleep(0.1)
    self.assertGreaterEqual(dist.value().mean, 100)

  def test_distribution_with_parameters(self):
    collection = monitoring.InMemoryMetricCollection('/test')
    dist = collection.get_distribution(
        'distribution', 'distribution description', {'field1': str}
    )
    self.assertEqual(dist.namespace, '/test')
    self.assertEqual(dist.name, 'distribution')
    self.assertEqual(dist.description, 'distribution description')
    self.assertEqual(dist.parameter_definitions, {'field1': str})
    self.assertEqual(dist.full_name, '/test/distribution')
    value = dist.value(field1='foo')
    self.assertEqual(value.count, 0)
    dist.record(1, field1='foo')
    dist.record(2, field1='foo')
    dist.record(3, field1='bar')
    value = dist.value(field1='foo')
    self.assertEqual(value.count, 2)
    value = dist.value(field1='bar')
    self.assertEqual(value.count, 1)

    dist = collection.get_distribution(
        'distribution2', 'distribution description', {'error': str}
    )
    with self.assertRaises(ValueError):
      with dist.record_duration():
        time.sleep(0.1)
        raise ValueError()
    self.assertGreaterEqual(dist.value(error='ValueError').mean, 100)
    with dist.record_duration():
      time.sleep(0.1)
    self.assertGreaterEqual(dist.value(error='').mean, 100)


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Html tooltip."""

from typing import Optional, Union
from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import object as pg_object
# pylint: disable=g-importing-member
from pyglove.core.views.html.base import Html
from pyglove.core.views.html.controls.base import HtmlControl
# pylint: disable=g-importing-member


@pg_object.use_init_args(
    ['content', 'for_element', 'id', 'css_classes', 'styles']
)
class Tooltip(HtmlControl):
  """A tooltip control.
  
  Attributes:
    content: The content of the tooltip. It could be a string or a HTML object.
    id: The id of the tooltip.
    css_classes: The CSS classes for the tooltip.
    for_element: The CSS selector for the element to attach the tooltip to.
      e.g. '.my-element' or '#my-element'.
  """

  content: Union[str, Html]
  for_element: Optional[str] = None

  def _to_html(self, **kwargs):
    if self.for_element is None:
      raise ValueError(
          'CSS selector `for_element` is required for tooltip to display.'
      )
    content = self.content
    if isinstance(self.content, str):
      content = Html.escape(self.content)
    return Html.element(
        'span',
        [content],
        id=self.element_id(),
        css_classes=[
            'tooltip',
            'html-content' if isinstance(content, Html) else None,
        ] + self.css_classes,
        styles=self.styles,
    ).add_style(
        """
        span.tooltip {
          visibility: hidden;
          white-space: pre-wrap;
          font-weight: normal;
          background-color: #484848;
          color: #fff;
          padding: 10px;
          border-radius: 6px;
          position: absolute;
          z-index: 1;
        }
        span.tooltip:hover {
          visibility: visible;
        }
        .tooltip.html-content {
          white-space: inherit;
          background-color: white;
          color: inherit;
          box-shadow: rgba(0, 0, 0, 0.16) 0px 1px 4px;
        }
        """,
        f"""
        {self.for_element}:hover + .tooltip {{
          visibility: visible;
        }}
        """
    )

  def update(self, content: Union[str, Html]) -> None:
    self._sync_members(content=self._update_content(content))
    if isinstance(content, Html):
      self._add_css_class('html-content')
    else:
      self._remove_css_class('html-content')


# Register converter for automatic conversion.
pg_typing.register_converter(str, Tooltip, Tooltip)
pg_typing.register_converter(Html, Tooltip, Tooltip)



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for formatting objects."""

import abc
import enum
import io
import sys
from typing import Any, Callable, ContextManager, Dict, List, Optional, Sequence, Set, Tuple
from pyglove.core.utils import thread_local


_TLS_STR_FORMAT_KWARGS = '_str_format_kwargs'
_TLS_REPR_FORMAT_KWARGS = '_repr_format_kwargs'


CustomFormatFn = Callable[
    [
        Any,    # Value to format.
        int,    # Root indent.
    ],
    # Returns a string or None. If None, the default `pg.format` will be used.
    Optional[str]
]


def str_format(**kwargs) -> ContextManager[Dict[str, Any]]:
  """Context manager for setting the default format kwargs for __str__."""
  return thread_local.thread_local_arg_scope(_TLS_STR_FORMAT_KWARGS, **kwargs)


def repr_format(**kwargs) -> ContextManager[Dict[str, Any]]:
  """Context manager for setting the default format kwargs for __repr__."""
  return thread_local.thread_local_arg_scope(_TLS_REPR_FORMAT_KWARGS, **kwargs)


class Formattable(metaclass=abc.ABCMeta):
  """Interface for classes whose instances can be pretty-formatted.

  This interface overrides the default ``__repr__`` and ``__str__`` method, thus
  all ``Formattable`` objects can be printed nicely.

  All symbolic types implement this interface.
  """

  # Additional format keyword arguments for `__str__`.
  __str_format_kwargs__ = dict(compact=False, verbose=True)

  # Additional format keyword arguments for `__repr__`.
  __repr_format_kwargs__ = dict(compact=True)

  @abc.abstractmethod
  def format(self,
             compact: bool = False,
             verbose: bool = True,
             root_indent: int = 0,
             **kwargs) -> str:
    """Formats this object into a string representation.

    Args:
      compact: If True, this object will be formatted into a single line.
      verbose: If True, this object will be formatted with verbosity.
        Subclasses should define `verbosity` on their own.
      root_indent: The start indent level for this object if the output is a
        multi-line string.
      **kwargs: Subclass specific keyword arguments.

    Returns:
      A string of formatted object.
    """

  def __str__(self) -> str:
    """Returns the full (maybe multi-line) representation of this object."""
    # NOTE(daiyip): we delegate the formatting logic to `pg.format` instead of
    # `Formattable.format` as `pg.format` could add common functionalities such
    # as `markdown` support.
    return format(self, **self.__str_kwargs__())

  def __str_kwargs__(self) -> Dict[str, Any]:
    """Returns the default format kwargs for __str__."""
    kwargs = dict(self.__str_format_kwargs__)
    kwargs.update(thread_local.thread_local_kwargs(_TLS_STR_FORMAT_KWARGS))
    return kwargs

  def __repr__(self) -> str:
    """Returns a single-line representation of this object."""
    # NOTE(daiyip): we delegate the formatting logic to `pg.format` instead of
    # `Formattable.format` as `pg.format` could add common functionalities such
    # as `markdown` support.
    return format(self, **self.__repr_kwargs__())

  def __repr_kwargs__(self) -> Dict[str, Any]:
    """Returns the default format kwargs for __repr__."""
    kwargs = dict(self.__repr_format_kwargs__)
    kwargs.update(thread_local.thread_local_kwargs(_TLS_REPR_FORMAT_KWARGS))
    return kwargs


class RawText(Formattable):
  """Raw text."""

  def __init__(self, text: str):
    self.text = text

  def format(self, *args, **kwargs):
    del args, kwargs
    return self.text

  def __eq__(self, other: Any) -> bool:
    if isinstance(other, RawText):
      return self.text == other.text
    elif isinstance(other, str):
      return self.text == other
    return False

  def __ne__(self, other: Any) -> bool:
    return not self.__eq__(other)


class BracketType(enum.IntEnum):
  """Bracket types used for complex type formatting."""
  # Round bracket.
  ROUND = 0

  # Square bracket.
  SQUARE = 1

  # Curly bracket.
  CURLY = 2


_BRACKET_CHARS = [
    ('(', ')'),
    ('[', ']'),
    ('{', '}'),
]


def bracket_chars(bracket_type: BracketType) -> Tuple[str, str]:
  """Gets bracket character."""
  return _BRACKET_CHARS[int(bracket_type)]


def kvlist_str(
    kvlist: List[Tuple[str, Any, Any]],
    compact: bool = True,
    verbose: bool = False,
    root_indent: int = 0,
    *,
    label: Optional[str] = None,
    bracket_type: BracketType = BracketType.ROUND,
    custom_format: Optional[CustomFormatFn] = None,
    memo: Optional[Set[int]] = None,
    **kwargs,
) -> str:
  """Formats a list key/value pairs into a comma delimited string.

  Args:
    kvlist: List of tuples in format of
      (key, value, default_value or a tuple of default values)
    compact: If True, format value in kvlist in compact form.
    verbose: If True, format value in kvlist in verbose.
    root_indent: The indent should be applied for values in kvlist if they are
      multi-line.
    label: (Optional) If not None, add label to brace all kv pairs.
    bracket_type: Bracket type used for embracing the kv pairs. Applicable only
      when `name` is not None.
    custom_format: An optional custom format function, which will be applied to
      each value (and child values) in kvlist. If the function returns None, it
      will fall back to the default `pg.format`.
    memo: A set of object ids that have been formatted. Used to avoid
      infinite recursion in the formatting process.
    **kwargs: Keyword arguments that will be passed through unto child
      ``Formattable`` objects.
  Returns:
    A formatted string from a list of key/value pairs delimited by comma.
  """
  s = io.StringIO()
  is_first = True
  bracket_start, bracket_end = bracket_chars(bracket_type)

  child_indent = (root_indent + 1) if label else root_indent
  body = io.StringIO()

  for k, v, d in kvlist:
    if isinstance(d, tuple):
      include_pair = True
      for sd in d:
        if sd == v:
          include_pair = False
          break
    else:
      include_pair = v != d
    if include_pair:
      if not is_first:
        body.write(',')
        body.write(' ' if compact else '\n')
      v = format(
          v, compact=compact,
          verbose=verbose,
          root_indent=child_indent,
          custom_format=custom_format,
          memo=memo,
          **kwargs
      )
      if not compact:
        body.write(_indent('', child_indent))
      if k:
        body.write(f'{k}={str_ext(v, custom_format, child_indent)}')
      else:
        body.write(str_ext(v, custom_format, child_indent))
      is_first = False

  if label and not is_first and not compact:
    body.write('\n')

  body = body.getvalue()
  if label is None:
    return body
  else:
    s.write(label)
    s.write(bracket_start)
    if body:
      if not compact:
        s.write('\n')
      s.write(body)
      if not compact:
        s.write(_indent('', root_indent))
    s.write(bracket_end)
    return s.getvalue()


def quote_if_str(value: Any) -> Any:
  """Quotes the value if it is a str."""
  if isinstance(value, str):
    return repr(value)
  return value


def comma_delimited_str(value_list: Sequence[Any]) -> str:
  """Gets comma delimited string."""
  return ', '.join(str(quote_if_str(v)) for v in value_list)


def auto_plural(
    number: int, singular: str, plural: Optional[str] = None) -> str:
  """Use singular form if number is 1, otherwise use plural form."""
  if plural is None:
    plural = singular + 's'
  return singular if number == 1 else plural


def format(   # pylint: disable=redefined-builtin
    value: Any,
    compact: bool = False,
    verbose: bool = True,
    root_indent: int = 0,
    list_wrap_threshold: int = 80,
    strip_object_id: bool = False,
    include_keys: Optional[Set[str]] = None,
    exclude_keys: Optional[Set[str]] = None,
    markdown: bool = False,
    max_str_len: Optional[int] = None,
    max_bytes_len: Optional[int] = None,
    *,
    custom_format: Optional[CustomFormatFn] = None,
    memo: Optional[Set[int]] = None,
    **kwargs,
) -> str:
  """Formats a (maybe) hierarchical value with flags.

  Args:
    value: The value to format.
    compact: If True, this object will be formatted into a single line.
    verbose: If True, this object will be formatted with verbosity. Subclasses
      should define `verbosity` on their own.
    root_indent: The start indent level for this object if the output is a
      multi-line string.
    list_wrap_threshold: A threshold in number of characters for wrapping a list
      value in a single line.
    strip_object_id: If True, format object as '<class-name>(...)' other than
      'object at <address>'.
    include_keys: A set of keys to include from the top-level dict or object.
    exclude_keys: A set of keys to exclude from the top-level dict or object.
      Applicable only when `include_keys` is set to None.
    markdown: If True, use markdown notion to quote the formatted object.
    max_str_len: The max length of the string to be formatted. If the string is
      longer than this length, it will be truncated.
    max_bytes_len: The max length of the bytes to be formatted. If the bytes is
      longer than this length, it will be truncated.
    custom_format: An optional custom format function, which will be applied to
      each value (and child values) in kvlist. If the function returns None, it
      will fall back to the default `pg.format`.
    memo: A set of object ids that have been formatted. Used to avoid
      infinite recursion in the formatting process.
    **kwargs: Keyword arguments that will be passed through unto child
      ``Formattable`` objects.

  Returns:
    A string representation for `value`.
  """
  # We allow custom_format to intercept the entire value if it's present.
  if custom_format is not None:
    result = custom_format(value, root_indent)
    if result is not None:
      return maybe_markdown_quote(result, markdown)

  if memo is None:
    memo = set()

  id_ = id(value)
  if id_ in memo:
    return f'<recursive {value.__class__.__name__} at 0x{id_:x}>'  # pylint: disable=bad-whitespace
  memo.add(id_)

  exclude_keys = exclude_keys or set()

  def _should_include_key(key: str) -> bool:
    if include_keys:
      return key in include_keys
    return key not in exclude_keys

  def _format_child(v):
    return format(
        v,
        compact=compact,
        verbose=verbose,
        root_indent=root_indent + 1,
        list_wrap_threshold=list_wrap_threshold,
        strip_object_id=strip_object_id,
        max_str_len=max_str_len,
        max_bytes_len=max_bytes_len,
        custom_format=custom_format,
        memo=memo,
        **kwargs
    )

  # `markdown` is only applied at the outter most level,
  # so we disable markdown format in the sub-tree for `str` and `repr`.
  with str_format(markdown=False), repr_format(markdown=False):
    if isinstance(value, Formattable):
      s = value.format(
          compact=compact,
          verbose=verbose,
          root_indent=root_indent,
          list_wrap_threshold=list_wrap_threshold,
          strip_object_id=strip_object_id,
          include_keys=include_keys,
          exclude_keys=exclude_keys,
          max_str_len=max_str_len,
          max_bytes_len=max_bytes_len,
          custom_format=custom_format,
          memo=memo,
          **kwargs
      )
    elif isinstance(value, (list, tuple)):
      # Always try compact representation if length is not too long.
      open_bracket, close_bracket = bracket_chars(
          BracketType.SQUARE if isinstance(value, list) else BracketType.ROUND)
      s = [open_bracket]
      s.append(', '.join([_format_child(elem) for elem in value]))
      s.append(close_bracket)
      s = [''.join(s)]
      if not compact and len(s[-1]) > list_wrap_threshold:
        s = [f'{open_bracket}\n']
        s.append(',\n'.join([
            _indent(_format_child(elem), root_indent + 1)
            for elem in value
        ]))
        s.append('\n')
        s.append(_indent(close_bracket, root_indent))
    elif isinstance(value, dict):
      if compact or not value:
        s = ['{']
        s.append(', '.join([
            f'{k!r}: {_format_child(v)}'
            for k, v in value.items() if _should_include_key(k)
        ]))
        s.append('}')
      else:
        s = ['{\n']
        s.append(',\n'.join([
            _indent(f'{k!r}: {_format_child(v)}', root_indent + 1)
            for k, v in value.items() if _should_include_key(k)
        ]))
        s.append('\n')
        s.append(_indent('}', root_indent))
    else:
      if isinstance(value, str):
        if max_str_len is not None and len(value) > max_str_len:
          value = value[:max_str_len] + '...'
        s = [repr_ext(value, custom_format, root_indent)]
      elif isinstance(value, bytes):
        if max_bytes_len is not None and len(value) > max_bytes_len:
          value = value[:max_bytes_len] + b'...'
        s = [repr_ext(value, custom_format, root_indent)]
      else:
        s = [repr_ext(value, custom_format, root_indent)
             if compact else str_ext(value, custom_format, root_indent)]
        if strip_object_id and 'object at 0x' in s[-1]:
          s = [f'{value.__class__.__name__}(...)']
  memo.remove(id_)
  return maybe_markdown_quote(''.join(s), markdown)


def _maybe_custom_format(
    v: Any,
    default_fn: Callable[[Any], str],
    custom_format: Optional[CustomFormatFn],
    root_indent: int,
) -> str:
  if custom_format is None:
    return default_fn(v)
  x = custom_format(v, root_indent)
  if x is None:
    return default_fn(v)
  return x


def str_ext(
    v: Any,
    custom_format: Optional[CustomFormatFn] = None,
    root_indent: int = 0,
) -> str:
  """"str operator with special format support."""
  return _maybe_custom_format(v, str, custom_format, root_indent)


def repr_ext(
    v: Any,
    custom_format: Optional[CustomFormatFn] = None,
    root_indent: int = 0,
) -> str:
  """repr operator with special format support."""
  return _maybe_custom_format(v, repr, custom_format, root_indent)


def maybe_markdown_quote(s: str, markdown: bool = True) -> str:
  """Maybe quote the formatted string with markdown."""
  if not markdown:
    return s
  if '\n' not in s:
    return f'`{s}`'
  else:
    return f'```\n{s}\n```'


def camel_to_snake(text: str, separator: str = '_') -> str:
  """Returns the snake case version of a camel case string."""
  chunks = []
  chunk_start = 0
  last_upper = 0
  length = len(text)
  for i, c in enumerate(text):
    if c.isupper():
      if last_upper < i - 1 or (i < length - 1 and text[i + 1].islower()):
        chunks.append(text[chunk_start:i])
        chunk_start = i
      last_upper = i
  chunks.append(text[chunk_start:])
  return (separator.join(c for c in chunks if c)).lower()


def printv(*args, **kwargs):
  """Prints formatted value."""
  fs = kwargs.pop('file', sys.stdout)
  print(*[format(v, **kwargs) for v in args], file=fs)


def _indent(text: str, indent: int) -> str:
  return ' ' * 2 * indent + text



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.geno.Space."""

import inspect
import random
import unittest

from pyglove.core import symbolic
from pyglove.core.geno.base import DNA
from pyglove.core.geno.categorical import manyof
from pyglove.core.geno.categorical import oneof
from pyglove.core.geno.custom import custom
from pyglove.core.geno.numerical import floatv
from pyglove.core.geno.space import constant
from pyglove.core.geno.space import Space


class SpaceTest(unittest.TestCase):
  """Tests for `pg.geno.Space`."""

  def test_init(self):
    _ = Space([
        oneof([constant(), constant()]),
        manyof(2, [constant(), constant()]),
        floatv(0.1, 1.0),
        custom()
    ])

    # Okay: location and name are the same for a decision point.
    _ = Space([
        oneof([constant()], location='x', name='x'),
        manyof(2, [constant(), constant()], location='y')
    ])

    # Test decision point clash on the same name.
    with self.assertRaisesRegex(
        ValueError, 'Found 2 decision point definitions clash on name'):
      _ = Space([
          oneof([constant()], name='x'),
          oneof([
              Space([
                  floatv(0., 1., name='x')
              ])])
      ])

    # Test decision points clash between name and id.
    with self.assertRaisesRegex(
        ValueError,
        'Found 2 decision point definitions clash between name .* and id .* '):
      _ = Space([
          oneof([constant()], name='x'),
          oneof([constant()], location='x')
      ])

    with self.assertRaisesRegex(TypeError, 'Expect .* but encountered .*'):
      _ = Space([constant()])

  def test_basics(self):
    space = constant()
    self.assertTrue(space.is_space)
    self.assertFalse(space.is_categorical)
    self.assertFalse(space.is_subchoice)
    self.assertFalse(space.is_numerical)
    self.assertFalse(space.is_custom_decision_point)

  def test_is_constant(self):
    self.assertTrue(constant().is_constant)
    self.assertFalse(Space([oneof([constant(), constant()])]).is_constant)
    self.assertFalse(Space([floatv(0.1, 0.2)]).is_constant)
    self.assertFalse(Space([custom()]).is_constant)

  def test_space_size(self):
    self.assertEqual(constant().space_size, 1)
    a = oneof([constant(), constant()])
    b = manyof(2, [constant(), constant(), constant()])
    c = floatv(-1.0, 1.0)
    d = custom()
    self.assertEqual(Space([a]).space_size, a.space_size)
    self.assertEqual(Space([b]).space_size, b.space_size)
    self.assertEqual(Space([c]).space_size, c.space_size)
    self.assertEqual(Space([d]).space_size, d.space_size)
    self.assertEqual(Space([a, b]).space_size, a.space_size * b.space_size)
    self.assertEqual(Space([a, b, c]).space_size, -1)
    self.assertEqual(Space([a, b, d]).space_size, -1)

  def test_len(self):
    self.assertEqual(len(constant()), 0)
    self.assertEqual(len(Space([
        oneof([
            oneof([constant(), constant()]),
            floatv(-1.0, 1.0)
        ]),
        manyof(2, [constant(), constant()]),
        floatv(-1.0, 1.0)
    ])), 6)

  def test_id(self):
    x = oneof([constant(), constant()])
    y = manyof(2, [constant(), constant()])
    z = floatv(-1.0, 1.0)
    a = oneof([x, y, z], location='a')
    b = manyof(2, [constant(), constant()], location='b')
    c = floatv(-1.0, 1.0, location='c')
    d = custom(location='d')
    space = Space([a, b, c, d], location='root')

    self.assertEqual(space.id, 'root')
    self.assertEqual(a.id, 'root.a')
    self.assertEqual(x.id, 'root.a[=0/3]')
    self.assertEqual(y.id, 'root.a[=1/3]')
    self.assertEqual(z.id, 'root.a[=2/3]')
    self.assertEqual(b.id, 'root.b')
    self.assertEqual(c.id, 'root.c')
    self.assertEqual(d.id, 'root.d')

  def test_get(self):
    x = oneof([constant(), constant()])
    y = manyof(2, [constant(), constant()], name='y')
    z = floatv(-1.0, 1.0)
    a = oneof([x, y, z], location='a')
    b = manyof(2, [constant(), constant()], location='b')
    c = floatv(-1.0, 1.0, location='c')
    d = custom(location='d', name='d')
    space = Space([a, b, c, d], location='root')

    # Get by ID.
    self.assertIs(space.get('root.a'), a)
    self.assertIs(space.get('root.a[=0/3]'), x)
    self.assertEqual(
        space.get('root.a[=1/3]'), [y.subchoice(0), y.subchoice(1)])
    self.assertIs(space.get('root.a[=2/3]'), z)
    self.assertEqual(space.get('root.b'), [b.subchoice(0), b.subchoice(1)])
    self.assertIs(space.get('root.c'), c)
    self.assertIs(space.get('root.d'), d)

    # Get by name.
    self.assertIsNone(space.get('a'))  # `a` does not specify a name.
    self.assertEqual(space.get('y'), [y.subchoice(0), y.subchoice(1)])
    self.assertIs(space.get('d'), d)

  def test_getitem(self):
    x = oneof([constant(), constant()])
    y = manyof(2, [constant(), constant()], name='y')
    z = floatv(-1.0, 1.0)
    a = oneof([x, y, z], location='a')
    b = manyof(2, [constant(), constant()], location='b')
    c = floatv(-1.0, 1.0, location='c')
    d = custom(location='d', name='d')
    space = Space([a, b, c, d], location='root')

    # Get by index and slice.
    self.assertIs(space[0], a)
    self.assertEqual(space[1:], [b, c, d])

    # Get by ID.
    self.assertIs(space['root.c'], c)

    # Get by name.
    self.assertEqual(space['y'], [y.subchoice(0), y.subchoice(1)])
    with self.assertRaises(KeyError):
      _ = space['a']    # `a` does not specify a name.

  def test_decision_ids(self):
    x = oneof([constant(), constant()])
    y = manyof(2, [constant(), constant()], name='y')
    z = floatv(-1.0, 1.0)
    a = oneof([x, y, z], location='a')
    b = manyof(2, [constant(), constant()], location='b')
    c = floatv(-1.0, 1.0, location='c')
    d = custom(location='d', name='d')
    space = Space([a, b, c, d], location='root')

    self.assertEqual(
        space.decision_ids,
        [
            'root.a',
            'root.a[=0/3]',
            'root.a[=1/3]',
            'root.a[=2/3]',
            'root.b',
            'root.c',
            'root.d',
        ])

  def test_decision_points(self):
    x = oneof([constant(), constant()])
    y = manyof(2, [constant(), constant()], name='y')
    z = floatv(-1.0, 1.0)
    a = oneof([x, y, z], location='a')
    b = manyof(2, [constant(), constant()], location='b')
    c = floatv(-1.0, 1.0, location='c')
    d = custom(location='d', name='d')
    space = Space([a, b, c, d], location='root')

    self.assertEqual(
        space.decision_points,
        [
            a, x, y.subchoice(0), y.subchoice(1), z,
            b.subchoice(0), b.subchoice(1), c, d,
        ])

  def test_named_decision_points(self):
    x = oneof([constant(), constant()])
    y = manyof(2, [constant(), constant()], name='y')
    z = floatv(-1.0, 1.0)
    a = oneof([x, y, z], location='a')
    b = manyof(2, [constant(), constant()], location='b')
    c = floatv(-1.0, 1.0, location='c')
    d = custom(location='d', name='d')
    space = Space([a, b, c, d], location='root')
    self.assertEqual(
        space.named_decision_points,
        {'d': d, 'y': [y.subchoice(0), y.subchoice(1)]})

  def test_set_get_userdata(self):
    s = constant()
    s.set_userdata('mydata', 1)
    self.assertEqual(s.userdata.mydata, 1)
    self.assertEqual(s.userdata['mydata'], 1)
    with self.assertRaises(KeyError):
      _ = s.userdata.nonexist_data
    with self.assertRaises(KeyError):
      _ = s.userdata['nonexist_data']

  def test_parent_spec(self):
    a = Space([                                   # a
        oneof([                                   # b
            oneof([                               # x
                constant(),
                constant()
            ]),
            manyof(2, [                           # y
                constant(),
                constant(),
                constant()
            ]),
            custom(),                             # z
        ]),
        floatv(min_value=0.0, max_value=1.0),     # c
        custom(),                                 # d
    ])
    b = a.elements[0]
    c = a.elements[1]
    d = a.elements[2]
    x = b.candidates[0].elements[0]
    y = b.candidates[1].elements[0]
    z = b.candidates[2].elements[0]

    self.assertIsNone(a.parent_spec)
    self.assertIs(b.parent_spec, a)
    self.assertIs(b.candidates[0].parent_spec, b)
    self.assertIs(x.parent_spec, b.candidates[0])
    self.assertIs(y.parent_spec, b.candidates[1])
    self.assertIs(z.parent_spec, b.candidates[2])
    self.assertIs(c.parent_spec, a)
    self.assertIs(d.parent_spec, a)

  def test_parent_choice(self):
    a = Space([                                # a
        oneof([                                # b
            oneof([                            # x
                constant(),
                constant()
            ]),
            manyof(2, [                        # y
                constant(),
                constant(),
                constant()
            ]),
            custom(),                          # z
        ]),
        floatv(min_value=0.0, max_value=1.0),  # c
        custom(),                              # d
    ])
    b = a.elements[0]
    c = a.elements[1]
    d = a.elements[2]
    x = b.candidates[0].elements[0]
    y = b.candidates[1].elements[0]
    z = b.candidates[2].elements[0]

    self.assertIsNone(a.parent_choice)
    self.assertIsNone(b.parent_choice)
    self.assertIsNone(b.subchoice(0).parent_choice)
    self.assertIs(x.parent_choice, b)
    self.assertIs(x.subchoice(0).parent_choice, b)
    self.assertIs(y.parent_choice, b)
    self.assertIs(y.subchoice(0).parent_choice, b)
    self.assertIs(y.subchoice(1).parent_choice, b)
    self.assertIs(z.parent_choice, b)
    self.assertIs(z.parent_choice, b)
    self.assertIsNone(c.parent_choice)
    self.assertIsNone(d.parent_choice)

  def test_validate(self):
    space = constant()
    space.validate(DNA(None))

    with self.assertRaisesRegex(ValueError, 'Extra DNA values encountered'):
      space.validate(DNA(1))

    space = Space([
        oneof([
            oneof([
                constant(),
                constant()
            ]),
            manyof(2, [
                constant(),
                constant(),
                constant()
            ]),
            constant(),
            custom(),
            floatv(min_value=-1.0, max_value=1.0)
        ]),
        floatv(min_value=0.0, max_value=1.0),
    ])
    space.validate(DNA([(0, 1), 0.5]))
    space.validate(DNA([(1, [0, 1]), 0.5]))
    space.validate(DNA([(3, 'abc'), 0.1]))
    space.validate(DNA([(4, 0.0), 0.1]))

    with self.assertRaisesRegex(
        ValueError,
        'Number of child values in DNA \\(.*\\) does not match '
        'the number of elements \\(.*\\)'):
      space.validate(DNA(None))

    with self.assertRaisesRegex(
        ValueError,
        'Number of child values in DNA \\(.*\\) does not match '
        'the number of elements \\(.*\\)'):
      space.validate(DNA(1))

    with self.assertRaisesRegex(
        ValueError, 'No child DNA provided for child space'):
      space.validate(DNA([0, 0.5]))

    with self.assertRaisesRegex(ValueError, 'Expect float value'):
      space.validate(DNA([2, 0]))

    with self.assertRaisesRegex(
        ValueError, 'DNA value should be no greater than '):
      space.validate(DNA([2, 1.5]))

    with self.assertRaisesRegex(
        ValueError, 'CustomDecisionPoint expects string type DNA'):
      space.validate(DNA([(3, 0.2), 0.1]))

  def test_first_dna(self):
    self.assertEqual(constant().first_dna(), DNA(None))

    space = Space([
        oneof([
            oneof([
                constant(),
                constant()
            ]),
            manyof(2, [
                constant(),
                constant(),
                constant()
            ])
        ]),
        floatv(min_value=0.0, max_value=1.0),
    ])
    self.assertEqual(space.first_dna(), DNA([(0, 0), 0.0]))

  def test_next_dna(self):
    space = constant()
    self.assertIsNone(space.next_dna(space.first_dna()))

    space = Space([floatv(-1.0, 1.0)])
    self.assertEqual(space.next_dna(), DNA(-1.0))
    with self.assertRaisesRegex(
        NotImplementedError, '`next_dna` is not supported on `Float` yet.'):
      space.next_dna(DNA(-1.0))

    space = Space([custom()])
    with self.assertRaisesRegex(
        NotImplementedError,
        '`next_dna` is not supported on \'CustomDecisionPoint\'.'):
      space.next_dna()

    space = Space([
        oneof([
            oneof([
                constant(),
                constant()
            ]),
            manyof(2, [
                constant(),
                constant(),
                constant()
            ]),
        ]),
        oneof([constant(), constant()])
    ])
    self.assertEqual(space.next_dna(), DNA([(0, 0), 0]))
    self.assertEqual(
        space.next_dna(DNA([(0, 1), 0])), DNA([(0, 1), 1]))
    self.assertEqual(
        space.next_dna(DNA([(0, 1), 1])), DNA([(1, [0, 1]), 0]))
    self.assertEqual(
        space.next_dna(DNA([(1, [0, 2]), 1])), DNA([(1, [1, 0]), 0]))

  def test_iter_dna(self):
    self.assertEqual(list(constant().iter_dna()), [DNA(None)])
    space = Space([
        oneof([
            oneof([
                constant(),
                constant()
            ]),
            manyof(2, [
                constant(),
                constant(),
                constant()
            ]),
        ]),
        oneof([constant(), constant()])
    ])
    self.assertEqual(list(space.iter_dna()), [
        DNA([(0, 0), 0]),
        DNA([(0, 0), 1]),
        DNA([(0, 1), 0]),
        DNA([(0, 1), 1]),
        DNA([(1, [0, 1]), 0]),
        DNA([(1, [0, 1]), 1]),
        DNA([(1, [0, 2]), 0]),
        DNA([(1, [0, 2]), 1]),
        DNA([(1, [1, 0]), 0]),
        DNA([(1, [1, 0]), 1]),
        DNA([(1, [1, 2]), 0]),
        DNA([(1, [1, 2]), 1]),
        DNA([(1, [2, 0]), 0]),
        DNA([(1, [2, 0]), 1]),
        DNA([(1, [2, 1]), 0]),
        DNA([(1, [2, 1]), 1])
    ])
    self.assertEqual(list(space.iter_dna(DNA([(1, [1, 2]), 1]))), [
        DNA([(1, [2, 0]), 0]),
        DNA([(1, [2, 0]), 1]),
        DNA([(1, [2, 1]), 0]),
        DNA([(1, [2, 1]), 1])
    ])

  def test_random_dna(self):
    r = random.Random(1)
    self.assertEqual(constant().random_dna(r), DNA(None))
    space = Space([
        oneof([
            oneof([
                constant(),
                constant()
            ]),
            manyof(2, [
                constant(),
                constant(),
                constant()
            ]),
        ]),
        oneof([constant(), constant()])
    ])
    self.assertEqual(space.random_dna(r), DNA([(0, 0), 1]))

  def test_inspection(self):
    space = Space([
        oneof([
            oneof([constant(), constant()]),
            constant()
        ]),
        floatv(min_value=0.0, max_value=1.0)
    ], location='a')

    # Test compact version, which overrides symbolic.Object.format.
    self.assertEqual(
        str(space),
        inspect.cleandoc("""Space({
                 0 = \'\': Choices(num_choices=1, [
                   (0): Space({
                     0 = \'\': Choices(num_choices=1, [
                       (0): Space()
                       (1): Space()
                     ])
                   })
                   (1): Space()
                 ])
                 1 = \'\': Float(min_value=0.0, max_value=1.0)
               })"""))

    # Test non-compact version.
    self.assertEqual(
        space.format(compact=False, verbose=True),
        symbolic.Object.format(space, compact=False, verbose=True))

  def test_serialization(self):
    space = Space([
        oneof([constant(), constant()], hints=1),
        manyof(2, [constant(), constant()]),
        floatv(0., 1.)
    ])
    json_dict = space.to_json()
    json_dict['userdata'] = None
    self.assertTrue(symbolic.eq(symbolic.from_json(json_dict), space))


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolic typing.

Overview
--------

To enable symbolic programmability on classes and functions, PyGlove intercepts
the assignment operation on the attributes of symbolic objects, to achieve two
goals:

  * Enable automatic type checking, value validation, conversion and
    transformation on attribute values. See `Runtime typing`_ for more details.

  * Allow symbolic attributes to be placeheld by special symbols, in order to
    represent abstract concepts such as a space of objects. E.g. hyper
    primitives like :func:`pyglove.oneof` placeholds a symbolic attribute to
    create a space of values for that attribute. See `Symbolic placeholding`_
    for more details.


Runtime typing
**************

Symbolic objects are intended to be manipulated after creation. Without a
runtime  typing system, things can go wrong easily. For instance, an `int`
attribute which was mistakenly modified at early program stages can be very
difficut to debug at later stages. PyGlove introduces a runtime type system
that automatically validates symbolic objects upon creation and modification,
minimizing boilerplated code for input validation, so the developer can focus on
the main business logic.

Understanding Schema
^^^^^^^^^^^^^^^^^^^^

PyGlove's runtime type system is based on the concept of ``Schema`` (
class :class:`pyglove.Schema`), which defines what symbolic attributes are held
by a symbolic type (e.g. a symbolic dict, a symbolic list or a symbolic class)
and what values each attribute accepts. A ``Schema`` object consists of a list
of ``Field`` (class :class:`pyglove.Field`), which define the acceptable
keys (class :class:`pyglove.KeySpec`) and their values (class
:class:`pyglove.ValueSpec`) for these types. A ``Schema`` object is usually
created automatically and associated with a symbolic type upon its declaration,
through decorators such as :func:`pyglove.members`, :func:`pyglove.symbolize` or
:func:`pyglove.functor`. For example::

  @pg.members([
      ('x', pg.typing.Int(default=1)),
      ('y', pg.typing.Float().noneable())
  ])
  class A(pg.Object):
    pass

  print(A.__schema__)

  @pg.symbolize([
      ('a', pg.typing.Int()),
      ('b', pg.typing.Float())
  ])
  def foo(a, b):
    return a + b

  print(foo.__schema__)


The first argument of all the decorators takes a list of field definitions,
with each described by a tuple of 4 items::

    (key specification, value specification, doc string, field metadata)

The **key specification** and **value specification** are required, while the
doc string and the field metadata are optional. The key specification defines
acceptable identifiers for this field, and the value specification defines the
attribute's value type, its default value, validation rules. The doc string will
serve as the description for the field, and the field metadata can be used for
field-based code generation.

The following code snippet illustrates all supported ``KeySpec`` and
``ValueSpec`` subclasses and their usage with a manually created schema::

    schema = pg.typing.create_schema([
        # Primitive types.
        ('a', pg.typing.Bool(default=True).noneable()),
        ('b', True),       # Equivalent to ('b', pg.typing.Bool(default=True)).
        ('c', pg.typing.Int()),
        ('d', 0),          # Equivalent to ('d', pg.typing.Int(default=0)).
        ('e', pg.typing.Int(
            min_value=0,
            max_value=10).noneable()),
        ('f', pg.typing.Float()),
        ('g', 1.0),        # Equivalent to ('g', pg.typing.Float(default=0.0)).
        ('h', pg.typing.Str()),
        ('i', 'foo'),      # Equivalent to ('i', pg.typing.Str(default='foo').
        ('j', pg.typing.Str(regex='foo.*')),

        # Enum type.
        ('l', pg.typing.Enum('foo', ['foo', 'bar', 0, 1]))

        # List type.
        ('m', pg.typing.List(pg.typing.Int(), size=2, default=[])),
        ('n', pg.typing.List(pg.typing.Dict([
            ('n1', pg.typing.List(pg.typing.Int())),
            ('n2', pg.typing.Str().noneable())
        ]), min_size=1, max_size=10, default=[])),

        # Dict type.
        ('o', pg.typing.Dict([
            ('o1', pg.typing.Int()),
            ('o2', pg.typing.List(pg.typing.Dict([
                ('o21', 1),
                ('o22', 1.0),
            ]))),
            ('o3', pg.typing.Dict([
                # Use of regex key,
                (pg.typing.StrKey('n3.*'), pg.typing.Int())
            ]))
        ]))

        # Tuple type.
        ('p', pg.typing.Tuple([
            ('p1', pg.typing.Int()),
            ('p2', pg.typing.Str())
        ]))

        # Object type.
        ('q', pg.typing.Object(A, default=A()))

        # Type type.
        ('r', pg.typing.Type(int))

        # Callable type.
        ('s', pg.typing.Callable([pg.typing.Int(), pg.typing.Int()],
                                  kw=[('a', pg.typing.Str())])),

        # Functor type (same as Callable, but only for symbolic.Functor).
        ('t', pg.typing.Functor([pg.typing.Str()],
                                 kwargs=[('a', pg.typing.Str())]))

        # Union type.
        ('u', pg.typing.Union([
            pg.typing.Int(),
            pg.typing.Str()
        ], default=1),

        # Any type.
        ('v', pg.typing.Any(default=1))
    ])


Schema inheritance
''''''''''''''''''

Symbolic attributes can be inherited during subclassing. Accordingly, the
schema that defines a symbolic class' attributes can be inherited too by its
subclasses. The fields from the bases' schema will be carried over into the
subclasses' schema, while the subclass can override, by redefining that field
with the same key. The subclass cannot override its base classes' field with
arbitrary value specs, it must be overriding non-frozen fields with more
restrictive validation rules of the same type, or change their default values.
See :meth:`pyglove.ValueSpec.extend` for more details.

The code snippet below illustrates schema inheritance during subclassing::

  @pg.members([
      ('x', pg.typing.Int(min_value=1)),
      ('y', pg.typing.Float()),
  ])
  class A(pg.Object):
    pass

  @pg.members([
      # Further restrict inherited 'x' by specifying the max value, as well
      # as providing a default value.
      ('x', pg.typing.Int(max_value=5, default=2)),
      ('z', pg.typing.Str('foo').freeze())
  ])
  class B(A):
    pass

  assert B.__schema__.fields.keys() == ['x', 'y', 'z']

  @pg.members([
      # Raises: 'z' is frozen in class B and cannot be extended further.
      ('z', pg.typing.Str())
  ])
  class C(B):
    pass


Automatic type conversions
^^^^^^^^^^^^^^^^^^^^^^^^^^

Type conversion is another mechanism to extend the typing system, which allows
the user to register converters between types. So when a value is assigned to a
attribute whose value specification does not match with the input value type, a
conversion will take place automatically if there is a converter from the input
value type to the type required by the value specification. For example::

  class A:
    def __init__(self, str):
      self._str = str

    def __str__(self):
      return self._str

    def __eq__(self, other):
      return isinstance(other, self.__class__) and self._str == other._str

  pg.typing.register_converter(A, str, str)
  pg.typing.register_converter(str, A, A)

  assert pg.typing.Str().accept(A('abc')) == 'abc'
  assert pg.typing.Object(A).accept('abc') == A('abc')


See :func:`pyglove.typing.register_converter` for more details.


Symbolic placeholding
*********************

Symbolic placeholding enables scenarios that requires an abstract (non-material)
representation of objects - for example - a search space or a partially bound
object. Such objects with placeheld attributes can only be used symbolically,
which means they can be symbolically queried or manipulated, but not yet
materialized to execute the business logic. For example::

  @pg.functor()
  def foo(x, y):
    return x + y

  f = f(x=pg.oneof(range(5)), y=pg.floatv(0.0, 2.0))

  # Error: `x` and `y` are not materialized, thus the functor cannot be
  # executed.
  f()

  # But we can symbolically manipulated it:
  f.rebind(x=1, y=2)

  # Returns 3
  f()


Symbolic placeholding is achieved by the :class:`pyglove.CustomTyping`
interface, which was intended as a mechanism to extend the typing system
horizontally without modifying the :class:`pyglove.ValueSpec` of existing
classes' attributes.

This is done by allowing the `CustomTyping` subclasses to take over value
validation and transformation from an attribute's original ``ValueSpec``
via :meth:`pyglove.CustomTyping.custom_apply` method.

Following is an example of using `CustomTyping` to extend the schema system::

  class FloatTensor(pg.typing.CustomTyping):

    def __init__(self, tensor):
      self._tensor = tensor

    def custom_apply(
        self, path, value_spec, allow_partial, child_transform):
      if isinstane(value_spec, pg.typing.Float):
        # Validate initial tensor, we can also add an TF operator to guard
        # tensor value to go beyond value_spec.min_value and max_value.
        value_spec.apply(self._tensor.numpy())
        # Shortcircuit .apply and returns object itself as final value.
        return (False, self)
      else:
        raise ValueError('FloatTensor can only be applied to float type.')

  @pg.members([
      ('x', pg.typing.Float())
  ])
  class Foo(pg.Object):
    pass

  # FloatTensor can be accepted for all symbolic attributes
  # with Float value spec.
  f = Foo(x=FloatTensor(tf.constant(1.0)))
"""

# pylint: disable=g-bad-import-order
# pylint: disable=g-importing-member
# pylint: disable=g-import-not-at-top

# Missing value.
from pyglove.core.typing.typed_missing import MISSING_VALUE  # Non-typed.
from pyglove.core.typing.typed_missing import MissingValue   # Typed.

# Class schema.
from pyglove.core.typing.class_schema import KeySpec
from pyglove.core.typing.class_schema import ValueSpec
from pyglove.core.typing.class_schema import Field
from pyglove.core.typing.class_schema import FieldKeyDef
from pyglove.core.typing.class_schema import FieldValueDef
from pyglove.core.typing.class_schema import FieldDef
from pyglove.core.typing.class_schema import Schema
from pyglove.core.typing.class_schema import create_field
from pyglove.core.typing.class_schema import create_schema
from pyglove.core.typing.class_schema import ForwardRef

# Concrete key specifications.
from pyglove.core.typing.key_specs import ConstStrKey
from pyglove.core.typing.key_specs import NonConstKey
from pyglove.core.typing.key_specs import StrKey
from pyglove.core.typing.key_specs import ListKey
from pyglove.core.typing.key_specs import TupleKey

# Concrete value specifications.
from pyglove.core.typing.value_specs import PrimitiveType
from pyglove.core.typing.value_specs import Bool
from pyglove.core.typing.value_specs import Str
from pyglove.core.typing.value_specs import Number
from pyglove.core.typing.value_specs import Int
from pyglove.core.typing.value_specs import Float
from pyglove.core.typing.value_specs import Enum
from pyglove.core.typing.value_specs import List
from pyglove.core.typing.value_specs import Tuple
from pyglove.core.typing.value_specs import Dict
from pyglove.core.typing.value_specs import Object
from pyglove.core.typing.value_specs import Callable
from pyglove.core.typing.value_specs import Functor
from pyglove.core.typing.value_specs import Type
from pyglove.core.typing.value_specs import Union
from pyglove.core.typing.value_specs import Any

from pyglove.core.typing.value_specs import ensure_value_spec

# Generic type aliases.
from pyglove.core.typing.value_specs import GenericMeta
from pyglove.core.typing.value_specs import Generic
from pyglove.core.typing.value_specs import GenericTypeAlias

from pyglove.core.typing.value_specs import Sequence
from pyglove.core.typing.value_specs import Optional

# `pg.typing.Any` is evaluated to `typing.Any` during type checking.
import typing as _typing
if _typing.TYPE_CHECKING:
  Any = _typing.Any

# Annotated.
from pyglove.core.typing.annotated import Annotated

# Type conversion.
from pyglove.core.typing.type_conversion import register_converter
from pyglove.core.typing.type_conversion import get_converter
from pyglove.core.typing.type_conversion import get_json_value_converter

# Inspect helpers.
from pyglove.core.typing.inspect import is_subclass
from pyglove.core.typing.inspect import is_instance
from pyglove.core.typing.inspect import get_outer_class
from pyglove.core.typing.inspect import get_type
from pyglove.core.typing.inspect import get_type_args
from pyglove.core.typing.inspect import is_generic
from pyglove.core.typing.inspect import has_generic_bases
from pyglove.core.typing.inspect import callable_eq

# Annotation conversion.
import pyglove.core.typing.annotation_conversion  # pylint: disable=unused-import

# Interface for custom typing.
from pyglove.core.typing.custom_typing import CustomTyping

# Annotation conversion
from pyglove.core.typing.annotation_conversion import annotation_from_str

# Callable signature.
from pyglove.core.typing.callable_signature import Argument
from pyglove.core.typing.callable_signature import CallableType
from pyglove.core.typing.callable_signature import Signature
from pyglove.core.typing.callable_signature import signature
from pyglove.core.typing.callable_signature import schema

# For backward compatibility.
get_signature = signature

# Callable extensions.
from pyglove.core.typing.callable_ext import PresetArgValue
from pyglove.core.typing.callable_ext import enable_preset_args
from pyglove.core.typing.callable_ext import preset_args
from pyglove.core.typing.callable_ext import CallableWithOptionalKeywordArgs

# JSON schema conversion.
from pyglove.core.typing.json_schema import to_json_schema

# PyType support.
from pyglove.core.typing.pytype_support import *

# pylint: enable=g-import-not-at-top
# pylint: enable=g-importing-member
# pylint: enable=g-bad-import-order



# Copyright 2019 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Setup for pip package."""

import collections
import datetime
import sys
from setuptools import find_namespace_packages
from setuptools import setup


def _get_version():
  """Gets current version of PyGlove package."""
  with open('pyglove/__init__.py') as fp:
    version = None
    for line in fp:
      if line.startswith('__version__'):
        g = {}
        exec(line, g)  # pylint: disable=exec-used
        version = g['__version__']
        break
  if version is None:
    raise ValueError('`__version__` not defined in `pyglove/__init__.py`')
  if '--nightly' in sys.argv:
    nightly_label = datetime.datetime.now().strftime('%Y%m%d%H%M')
    version = f'{version}.dev{nightly_label}'
    sys.argv.remove('--nightly')
  return version


def _parse_requirements(
    requirements_txt_path: str
) -> tuple[list[str], dict[str, list[str]]]:
  """Parses the require and extras_require for setup() from requirements.txt."""

  # Currently a requirements.txt is being used to specify dependencies. In order
  # to avoid specifying it in two places, we're going to use that file as the
  # source of truth.
  extras = collections.defaultdict(list)
  paths = ['require']

  def add_requirement(requirement: str, extra_key: str) -> None:
    if requirement not in extras[extra_key]:
      extras[extra_key].append(requirement)

  with open(requirements_txt_path) as file:
    for line in file:
      line = line.strip()
      if not line:
        continue

      if line.startswith('# extras:'):
        extra_path = line[line.find(':') + 1:].split('-')
        paths = [
            '-'.join(extra_path[:i + 1]) for i in range(len(extra_path))
        ]
      else:
        requirement, *_ = line.split('#')
        if requirement:
          for p in paths:
            add_requirement(requirement, p)
          add_requirement(requirement, 'all')

  require = extras.pop('require')
  return require, dict(extras)


_VERSION = _get_version()

install_requires, extras_require = _parse_requirements('requirements.txt')

setup(
    name='pyglove',
    version=_VERSION,
    url='https://github.com/google/pyglove',
    license='Apache License 2.0',
    author='PyGlove Authors',
    description='PyGlove: A library for manipulating Python objects.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author_email='pyglove-authors@google.com',
    # Contained modules and scripts.
    packages=find_namespace_packages(include=['pyglove*']),
    install_requires=install_requires,
    extras_require=extras_require,
    requires_python='>=3.9',
    include_package_data=True,
    # PyPI package information.
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries',
    ],
    keywords=(
        'ai machine learning automl mutable symbolic '
        'framework meta-programming'),
)
